@@ -1,4 +1,4 @@
-## ---- include = FALSE---------------------------------------------------------
+## ----include = FALSE----------------------------------------------------------
 knitr::opts_chunk$set(
   collapse = TRUE,
   comment = "#>"
@@ -24,13 +24,16 @@ m_df <- hgwr(
 )
 
 ## -----------------------------------------------------------------------------
-m_sf
+m_df
 
 ## -----------------------------------------------------------------------------
-summary(m_sf)
+summary(m_df)
 
 ## -----------------------------------------------------------------------------
-head(coef(m_sf))
-head(fitted(m_sf))
-head(residuals(m_sf))
+summary(m_df, test_hetero = T)
+
+## -----------------------------------------------------------------------------
+head(coef(m_df))
+head(fitted(m_df))
+head(residuals(m_df))
 
